﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Text.RegularExpressions;
using System.Globalization;
using System.IO;
using System.Drawing;

namespace Unclassified
{
	/// <summary>
	/// Provides easy read and write access to XML documents through XPath selectors.
	/// </summary>
	public class EasyXml
	{
		/// <summary>
		/// Internal XML document that is acted upon.
		/// </summary>
		private XmlDocument xd;

		/// <summary>
		/// Gets or sets the current XML document as a formatted string.
		/// </summary>
		public string Xml
		{
			get
			{
				using (MemoryStream ms = new MemoryStream())
				{
					XmlWriterSettings xws = new XmlWriterSettings();
					xws.Encoding = Encoding.UTF8;
					xws.Indent = true;
					xws.IndentChars = "  ";
					XmlWriter xw = XmlWriter.Create(ms, xws);
					xd.Save(xw);
					return Encoding.UTF8.GetString(ms.ToArray());
				}
			}
			set
			{
				xd.LoadXml(value);
			}
		}

		/// <summary>
		/// Gets the internal XML document for direct access.
		/// </summary>
		public XmlDocument XmlDocument
		{
			get
			{
				return xd;
			}
		}

		/// <summary>
		/// Initialises a new instance of the EasyXml class and created a new empty XML document.
		/// </summary>
		public EasyXml()
		{
			xd = new XmlDocument();
		}

		/// <summary>
		/// Loads an XML file into this document.
		/// </summary>
		/// <param name="fileName">XML file name</param>
		public void Load(string fileName)
		{
			xd.Load(fileName);
		}

		/// <summary>
		/// Saves this document to an XML file.
		/// </summary>
		/// <param name="fileName">XML file name</param>
		public void Save(string fileName)
		{
			XmlWriterSettings xws = new XmlWriterSettings();
			xws.Encoding = Encoding.UTF8;
			xws.Indent = true;
			xws.IndentChars = "\t";
			xws.OmitXmlDeclaration = false;

			XmlWriter xw = XmlWriter.Create(fileName, xws);
			xd.WriteTo(xw);
			xw.Close();
		}

		/// <summary>
		/// Determines whether an element or attribute specified by the XPath selector exists in the document.
		/// </summary>
		/// <param name="xPath">XPath selector to check</param>
		/// <returns>true if an element or attribute exists, false otherwise.</returns>
		public bool Exists(string xPath)
		{
			return Exists(xd, xPath);
		}

		/// <summary>
		/// Determines whether an element or attribute specified by the XPath selector exists in the document.
		/// </summary>
		/// <param name="baseNode">XML node to start searching at</param>
		/// <param name="xPath">XPath selector to check</param>
		/// <returns>true if an element or attribute exists, false otherwise.</returns>
		public bool Exists(XmlNode baseNode, string xPath)
		{
			return baseNode.SelectSingleNode(xPath) != null;
		}

		/// <summary>
		/// Reads a string value from a node.
		/// </summary>
		/// <param name="xPath">XPath selector for the node</param>
		/// <returns>Text contents or attribute value of the selected node.</returns>
		public string ReadString(string xPath)
		{
			return ReadString(xd, xPath, null);
		}

		/// <summary>
		/// Reads a string value from a node.
		/// </summary>
		/// <param name="xPath">XPath selector for the node</param>
		/// <param name="defaultValue">Value to be returned if the node does not exist</param>
		/// <returns>Text contents or attribute value of the selected node.</returns>
		public string ReadString(string xPath, string defaultValue)
		{
			return ReadString(xd, xPath, defaultValue);
		}

		public string ReadString(XmlNode baseNode, string xPath)
		{
			return ReadString(baseNode, xPath, null);
		}

		public string ReadString(XmlNode baseNode, string xPath, string defaultValue)
		{
			XmlNode node = baseNode.SelectSingleNode(xPath);
			XmlElement element = node as XmlElement;
			if (element != null)
			{
				return element.InnerText;
			}
			XmlAttribute attribute = node as XmlAttribute;
			if (attribute != null)
			{
				return attribute.Value;
			}
			if (defaultValue != null)
			{
				return defaultValue;
			}
			if (node != null)
			{
				throw new NotSupportedException("Unsupported node type selected.");
			}
			else
			{
				throw new NotSupportedException("Node not found.");
			}
		}

		public int ReadInt(string xPath)
		{
			return ReadInt(xd, xPath, null);
		}

		public int ReadInt(string xPath, int? defaultValue)
		{
			return ReadInt(xd, xPath, defaultValue);
		}

		public int ReadInt(XmlNode baseNode, string xPath)
		{
			return ReadInt(baseNode, xPath, null);
		}

		public int ReadInt(XmlNode baseNode, string xPath, int? defaultValue)
		{
			int i;
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			if (int.TryParse(s, NumberStyles.Integer, CultureInfo.InvariantCulture, out i))
			{
				return i;
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			throw new FormatException();
		}

		public decimal ReadDecimal(string xPath)
		{
			return ReadDecimal(xd, xPath, null);
		}

		public decimal ReadDecimal(string xPath, decimal? defaultValue)
		{
			return ReadDecimal(xd, xPath, defaultValue);
		}

		public decimal ReadDecimal(XmlNode baseNode, string xPath)
		{
			return ReadDecimal(baseNode, xPath, null);
		}

		public decimal ReadDecimal(XmlNode baseNode, string xPath, decimal? defaultValue)
		{
			decimal d;
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			if (decimal.TryParse(s, NumberStyles.Number, CultureInfo.InvariantCulture, out d))
			{
				return d;
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			throw new FormatException();
		}

		public bool ReadBool(string xPath)
		{
			return ReadBool(xd, xPath, null);
		}

		public bool ReadBool(string xPath, bool? defaultValue)
		{
			return ReadBool(xd, xPath, defaultValue);
		}

		public bool ReadBool(XmlNode baseNode, string xPath)
		{
			return ReadBool(baseNode, xPath, null);
		}

		public bool ReadBool(XmlNode baseNode, string xPath, bool? defaultValue)
		{
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			if (!string.IsNullOrEmpty(s))
			{
				return s == "true" || s == "1" || s == "yes" || s == "on";
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			return false;
		}

		public DateTime ReadDateTime(string xPath)
		{
			return ReadDateTime(xd, xPath, null);
		}

		public DateTime ReadDateTime(string xPath, DateTime? defaultValue)
		{
			return ReadDateTime(xd, xPath, defaultValue);
		}

		public DateTime ReadDateTime(XmlNode baseNode, string xPath)
		{
			return ReadDateTime(baseNode, xPath, null);
		}

		public DateTime ReadDateTime(XmlNode baseNode, string xPath, DateTime? defaultValue)
		{
			DateTime dt;
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			if (DateTime.TryParse(s, out dt))
			{
				return dt;
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			throw new FormatException();
		}

		public TimeSpan ReadTimeSpan(string xPath)
		{
			return ReadTimeSpan(xd, xPath, null);
		}

		public TimeSpan ReadTimeSpan(string xPath, TimeSpan? defaultValue)
		{
			return ReadTimeSpan(xd, xPath, defaultValue);
		}

		public TimeSpan ReadTimeSpan(XmlNode baseNode, string xPath)
		{
			return ReadTimeSpan(baseNode, xPath, null);
		}

		public TimeSpan ReadTimeSpan(XmlNode baseNode, string xPath, TimeSpan? defaultValue)
		{
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			Match m = Regex.Match(s, "^P((?<D>[0-9]+)D)?(T((?<H>[0-9]+)H)?((?<M>[0-9]+)M)?((?<S>[0-9]+)S)?)?$");
			if (m.Success)
			{
				int days = string.IsNullOrEmpty(m.Groups["D"].Value) ? 0 : int.Parse(m.Groups["D"].Value);
				int hours = string.IsNullOrEmpty(m.Groups["H"].Value) ? 0 : int.Parse(m.Groups["H"].Value);
				int mins = string.IsNullOrEmpty(m.Groups["M"].Value) ? 0 : int.Parse(m.Groups["M"].Value);
				int secs = string.IsNullOrEmpty(m.Groups["S"].Value) ? 0 : int.Parse(m.Groups["S"].Value);
				return TimeSpan.FromSeconds(days * 86400 + hours * 3600 + mins * 60 + secs);
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			throw new FormatException();
		}

		public Color ReadColor(string xPath)
		{
			return ReadColor(xd, xPath, null);
		}

		public Color ReadColor(string xPath, Color? defaultValue)
		{
			return ReadColor(xd, xPath, defaultValue);
		}

		public Color ReadColor(XmlNode baseNode, string xPath)
		{
			return ReadColor(baseNode, xPath, null);
		}

		public Color ReadColor(XmlNode baseNode, string xPath, Color? defaultValue)
		{
			string s = ReadString(baseNode, xPath, defaultValue.HasValue ? "" : null);
			Match m = Regex.Match(s, "^#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$", RegexOptions.IgnoreCase);
			if (m.Success)
			{
				int r = int.Parse(m.Groups[1].Value, NumberStyles.HexNumber);
				int g = int.Parse(m.Groups[2].Value, NumberStyles.HexNumber);
				int b = int.Parse(m.Groups[3].Value, NumberStyles.HexNumber);
				return Color.FromArgb(r, g, b);
			}
			if (defaultValue != null)
			{
				return defaultValue.Value;
			}
			throw new FormatException();
		}

		public XmlNode Create(string xPath)
		{
			return Create(xd, xPath, false);
		}

		public XmlNode Create(XmlNode baseNode, string xPath)
		{
			return Create(baseNode, xPath, false);
		}

		public XmlNode CreateNew(string xPath)
		{
			return Create(xd, xPath, true);
		}

		public XmlNode CreateNew(XmlNode baseNode, string xPath)
		{
			return Create(baseNode, xPath, true);
		}

		public XmlNode Create(XmlNode baseNode, string xPath, bool newNode)
		{
			// Idea (less strict parsing): xPath.Trim('/').Split('/')

			if (!Regex.IsMatch(xPath, "^(/?[-._a-z0-9]+)(/[-._a-z0-9]+)*(/@[-._a-z0-9]+)?$", RegexOptions.IgnoreCase))
			{
				throw new NotSupportedException("Unsupported XPath expression for creating an XML node.");
			}
			string[] elements = xPath.Split('/');

			if (newNode && elements[elements.Length - 1][0] == '@')
			{
				throw new NotSupportedException("You cannot create a new node with an XPath expression for an XML attribute.");
			}

			XmlNode node = baseNode;
			for (int i = 0; i < elements.Length; i++)
			{
				string element = elements[i];
				if (element.Length == 0) continue;

				XmlNode subNode = node.SelectSingleNode(element);
				bool skip = newNode && i == elements.Length - 1;   // Skip existing to create a new element
				if (subNode != null && !skip)
				{
					node = subNode;
				}
				else
				{
					if (element[0] == '@')
					{
						string attribute = element.Substring(1);
						subNode = xd.CreateAttribute(attribute);
						node.Attributes.Append((XmlAttribute) subNode);
						node = subNode;
					}
					else
					{
						subNode = xd.CreateElement(element);
						node.AppendChild(subNode);
						node = subNode;
					}
				}
			}
			return node;
		}

		public void Write(string xPath, int value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, int value)
		{
			Write(baseNode, xPath, value.ToString(CultureInfo.InvariantCulture));
		}

		public void Write(string xPath, decimal value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, decimal value)
		{
			Write(baseNode, xPath, value.ToString(CultureInfo.InvariantCulture));
		}

		public void Write(string xPath, bool value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, bool value)
		{
			Write(baseNode, xPath, value ? "true" : "false");
		}

		public void Write(string xPath, DateTime value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, DateTime value)
		{
			Write(baseNode, xPath, value.ToString("s"));
		}

		public void Write(string xPath, TimeSpan value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, TimeSpan value)
		{
			TimeSpan t = (TimeSpan) value;
			string s = "P";
			if (t.Days > 0) s += t.Days + "D";
			if (t.Hours > 0 || t.Minutes > 0 || t.Seconds > 0) s += "T";
			if (t.Hours > 0) s += t.Hours + "H";
			if (t.Minutes > 0) s += t.Minutes + "M";
			if (t.Seconds > 0) s += t.Seconds + "S";
			if (s == "P") s = "PT0S";
			Write(baseNode, xPath, s);
		}

		public void Write(string xPath, Color value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, Color value)
		{
			Write(baseNode, xPath, "#" + value.R.ToString("x2") + value.G.ToString("x2") + value.B.ToString("x2"));
		}

		public void Write(string xPath, string value)
		{
			Write(xd, xPath, value);
		}

		public void Write(XmlNode baseNode, string xPath, string value)
		{
			XmlNode node = baseNode.SelectSingleNode(xPath);
			if (node == null)
			{
				node = Create(baseNode, xPath);
			}

			XmlElement element = node as XmlElement;
			if (element != null)
			{
				element.InnerText = value;
				return;
			}
			XmlAttribute attribute = node as XmlAttribute;
			if (attribute != null)
			{
				attribute.Value = value;
				return;
			}
			throw new NotSupportedException("Unsupported node type selected.");
		}
	}
}
